unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, Jpeg, Math, ImgList;

type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte;
  end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;

  TfrmMain = class(TForm)
    imgView: TImage;
    dlgOpenImg: TOpenDialog;
    pnlBottom: TPanel;
    tb1: TTrackBar;
    tb2: TTrackBar;
    ilHandles: TImageList;
    tb3: TTrackBar;
    lbl1: TLabel;
    lbl2: TLabel;
    lbl3: TLabel;
    chk1: TCheckBox;
    lblTitle: TLabel;
    btnOpen: TButton;
    btnGo: TButton;
    dlgSaveImg: TSaveDialog;
    tb4: TTrackBar;
    lbl4: TLabel;
    procedure btnOpenClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure imgViewMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure imgViewMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure imgViewMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure imgViewMouseLeave(Sender: TObject);
    procedure tb2Change(Sender: TObject);
    procedure btnGoClick(Sender: TObject);
    procedure imgViewMouseEnter(Sender: TObject);
    procedure chk1Click(Sender: TObject);
  private
    procedure MakePreviewImage;
    procedure DoTiltShift(Bmp: TBitmap; Pt1, Pt2: TPoint; Render, Save: Boolean);
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

var
  TheImage, Img: TBitmap;
  LY, CP: Integer;
  PT1, PT2: TPoint;
  aR, aG, aB: array[0..255] of Byte;
  ImgLoaded: Boolean;
  DoNow: Boolean;


function GetAnglePt2FromPt1(pt1, pt2: TPoint): single;
// http://www.angusj.com/delphitips/beziertext.php
begin
  //nb: result is in radians
  dec(pt2.X,pt1.X);
  dec(pt2.Y,pt1.Y);
  with pt2 do
    if X = 0 then begin
      result := pi/2;
      if Y > 0 then result := 3*result; //nb: Y axis is +ve down
    end else begin
      result := arctan2(-Y,X);
      if result < 0 then result := result + pi * 2;
    end;
end;

function GetPtAtDistAndAngleFromPt(pt: TPoint;
  dist: integer; angle: single): TPoint;
// http://www.angusj.com/delphitips/beziertext.php
begin
  result.X := round(dist * cos(angle));
  result.Y := -round(dist * sin(angle)); //nb: Y axis is +ve down
  inc(result.X , pt.X);
  inc(result.Y , pt.Y);
end;

procedure ClipLine(var x1, y1, x2, y2: Integer; ClipRect: TRect);
// http://www.nondot.org/sabre/graphpro/line6.html
const
  CodeBottom = 1; CodeTop    = 2;             { BitFields for output codes }
  CodeLeft   = 4; CodeRight  = 8;

  function CompOutCode(x, y : Integer): Byte;  { Nested function }
  var
    Code : BYTE;
  begin
    Code := 0;
    if y > ClipRect.Bottom then Code:= CodeBottom else if y < ClipRect.Top then Code:= CodeTop;
    IF x > ClipRect.Right then Code:= Code+CodeRight else if X < ClipRect.Left then Code:= Code+CodeLeft;
    CompOutCode := Code;
  end;

var
  OutCode0,         { The code of the first endpoint  }
  OutCode1,         { The code of the second endpoint }
  OutCodeOut : Byte;
  x, y : Integer;
BEGIN
  x:= 0; y:= 0;
  OutCode0 := CompOutCode(x1, y1);            { Compute the original codes   }
  OutCode1 := CompOutCode(x2, y2);
  while (OutCode0 <> 0) or (OutCode1 <> 0) do begin{ While not Trivially Accepted }
    if (OutCode0 AND OutCode1) <> 0 then Exit      { Trivial Reject }
    else begin        { Failed both tests, so calculate the line segment to clip }
      if OutCode0 > 0 then OutCodeOut := OutCode0    { Clip the first point }
      else OutCodeOut := OutCode1;   { Clip the last point  }
      if (OutCodeOut and CodeBottom) = CodeBottom then begin { Clip the line to the bottom of the viewport     }
        Y := ClipRect.Bottom;
        X := x1 + LONGINT(x2 - x1) * LONGINT(y - y1) div (y2 - y1);
      end
      else if (OutCodeOut and CodeTop) = CodeTop then begin { Clip the line to the top of the viewport        }
        Y := ClipRect.Top;
        X := x1+LONGINT(x2 - x1) * LONGINT(y - y1) div (y2 - y1);
      end
      else if (OutCodeOut and CodeRight) = CodeRight then begin { Clip the line to the right edge of the viewport }
        X := ClipRect.Right;
        Y := y1+LONGINT(y2 - y1) * LONGINT(x - x1) div (x2 - x1);
      end
      else if (OutCodeOut and CodeLeft) = CodeLeft then begin { Clip the line to the left edge of the viewport  }
        X := ClipRect.Left;
        Y := y1+LONGINT(y2 - y1) * LONGINT(x - x1) div (x2 - x1);
      end;
      if (OutCodeOut = OutCode0) then begin       { Modify the first coordinate   }
        X1 := x; Y1 := y;                   { Update temporary variables    }
        OutCode0 := CompOutCode(x1, y1);    { Recalculate the OutCode       }
      end else begin                                  { Modify the second coordinate  }
        X2 := x; Y2 := y;                   { Update temporary variables    }
        OutCode1 := CompOutCode(x2, y2);    { Recalculate the OutCode       }
      end;
    end;
  end;
end;

function Distance(x1, y1, x2, y2: integer): integer;
var
  i, a, r0, r1: integer;
begin
  r0 := x2 - x1;
  r1 := y2 - y1;
  a := (r0 * r0) + (r1 * r1);
  r1 := (1 + a) div 2;
  for i := 1 to 12 do begin
    r0 := r1;
    if r0 > 0 then r1 := (r0 + (a div r0)) div 2;
  end;
  distance := r1;
end;

function CalcImgSize(w, h, tw, th: integer): TPoint;
begin
  Result:= Point(0, 0);
  if (w = 0) or (h = 0) then Exit;
  if (w < tw) and (h < th) then Result:= Point(w, h)
  else begin
    if w > h then begin
      if w < tw then tw := w;
      Result:= Point(tw, Trunc(tw * h / w));
      if Result.Y > th then Result:= Point(Trunc(th * w / h), th);
    end else begin
      if h < th then th := h;
      Result:= Point(Trunc(th * w / h), th);
      if Result.X > tw then Result:= Point(tw, Trunc(tw * h / w));
    end;
  end;
end;

procedure SmoothReSize(Bmp, Dest: tBitmap);
var
  x, y, R, G, B, PSrc1, PSrc2, SrcFill, PDst, DstFill: integer;
  Src : TBitmap;
begin
  Src:= TBitmap.Create;
  Src.PixelFormat:= pf24bit;
  Src.Width:= Dest.Width shl 1;
  Src.Height:= Dest.Height shl 1;
  Src.Canvas.StretchDraw(Src.Canvas.ClipRect, bmp);
  PSrc1:= Integer(Src.ScanLine[0]);
  PSrc2:= Integer(Src.ScanLine[1]);
  SrcFill := (PSrc2 - PSrc1);
  SrcFill := SrcFill + (SrcFill - Src.Width * 3);
  PDst := Integer(Dest.ScanLine[0]);
  DstFill := (Integer(Dest.ScanLine[1]) - PDst) - (Dest.Width * 3);
  for y:= 0 to Dest.Height - 1 do begin
    for x:= 0 to Dest.Width - 1 do begin
      B:= Pbyte(PSrc1)^;
      G:= PByte(PSrc1 + 1)^;
      R:= PByte(PSrc1 + 2)^;
      B:= B + PByte(PSrc1 + 3)^;
      G:= G + PByte(PSrc1 + 4)^;
      R:= R + PByte(PSrc1 + 5)^;
      B:= B + PByte(PSrc2)^;
      G:= G + PByte(PSrc2 + 1)^;
      R:= R + PByte(PSrc2 + 2)^;
      B:= B + PByte(PSrc2 + 3)^;
      G:= G + PByte(PSrc2 + 4)^;
      R:= R + PByte(PSrc2 + 5)^;
      Inc(PSrc1, 6);
      Inc(PSrc2, 6);
      PByte(PDst)^:= B shr 2;
      PByte(PDst + 1)^:= G shr 2;
      PByte(PDst + 2)^:= R shr 2;
      Inc(PDst, 3);
    end;
    Inc(PSrc1, SrcFill);
    Inc(PSrc2, SrcFill);
    Inc(PDst, DstFill);
  end;
  Src.Free;
end;

procedure fxBCS(Dest: TBitmap; B, C, S: integer);
// Brightness values [0..512] def 255
// Contrast values [-100..100] def 0
// Saturation values [0..512] def 255
var
  x, y, i, k, v, pv, ci1, ci2, ci3, alpha, s1, s2, SLDest, DestOff: Integer;
  a: double;
  p: PRGB24;
  color: TRGB24;
  BrightnesLut, ContrastLut, BCLut: array[0..255] of byte;
  SaturationLut: record
    Grays: array[0..767] of Integer; Alpha: array[Byte] of Word;
  end;
begin
  if c = 100 then c := 99;
  for i := 0 to 255 do begin
    if c > 0 then a := 1 / cos(c * (PI / 200))
    else          a := 1 * cos(c * (3.1416 / 200));
    v := Round(a * (i - 170) + 170);
    if v > 255 then v := 255 else if v < 0 then v := 0;
    ContrastLut[i] := v;
  end;
  for i := 0 to 255 do begin
    alpha := b;
    k := 256 - alpha;
    v := (k + alpha * i) div 256;
    if v < 0 then v := 0 else if v > 255 then v := 255;
    BCLut[i] := ContrastLut[v];
  end;
  x := 0;
  for i := 1 to 256 do SaturationLut.Alpha[i - 1] := (i * s) shr 8;
  for i := 0 to 255 do begin
    y := i - SaturationLut.Alpha[i];
    SaturationLut.Grays[x] := y; Inc(x);
    SaturationLut.Grays[x] := y; Inc(x);
    SaturationLut.Grays[x] := y; Inc(x);
  end;
  SLDest := Integer(dest.ScanLine[0]);
  DestOff := Integer(dest.ScanLine[1]) - SLDest;
  for y := 0 to Dest.Height - 1 do begin
    for x := 0 to Dest.Width - 1 do begin
      color := pRGBArray(SLDest)[x];
      v := color.R + color.G + color.B;
      ci1 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.B];
      ci2 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.G];
      ci3 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.R];
      if ci1 < 0 then ci1 := 0 else if ci1 > 255 then ci1 := 255;
      if ci2 < 0 then ci2 := 0 else if ci2 > 255 then ci2 := 255;
      if ci3 < 0 then ci3 := 0 else if ci3 > 255 then ci3 := 255;
      PRGBArray(SLDest)[x].B := BCLut[ci1];
      PRGBArray(SLDest)[x].G := BCLut[ci2];
      PRGBArray(SLDest)[x].R := BCLut[ci3];
    end;
    inc(SLDest, DestOff);
  end;
end;

procedure fxSplitBlur(Src: TBitmap; Amount: Integer);
var
  x, y, x1, x2, y1, y2, ptrS: integer;
  p: PRGB24;
  SCL: PRGBArray;
begin
  if Amount = 0 then Exit;
  SCL := src.ScanLine[0];
  ptrS := integer(src.ScanLine[1]) - integer(SCL);
  for y := 0 to src.Height - 1 do
  begin
    if y - amount < 0 then y1 := y else y1 := y - amount;
    if y + amount < src.Height then y2 := y + amount else y2 := src.Height - y;
    for x := 0 to src.Width - 1 do begin
      if x - amount < 0 then x1 := x else x1 := x - amount;
      if x + amount < src.Width then x2 := x + amount else x2 := src.Width - x;
      p := pRGB24(PtrS * y + (x * 3) + Integer(SCL));
      p.r := (pRGB24(PtrS * y1 + (x1 * 3) + Integer(SCL)).R +
        pRGB24(PtrS * y2 + (x1 * 3) + Integer(SCL)).R +
        pRGB24(PtrS * y1 + (x2 * 3) + Integer(SCL)).R +
        pRGB24(PtrS * y2 + (x2 * 3) + Integer(SCL)).R) shr 2;
      p.g := (pRGB24(PtrS * y1 + (x1 * 3) + Integer(SCL)).G +
        pRGB24(PtrS * y2 + (x1 * 3) + Integer(SCL)).G +
        pRGB24(PtrS * y1 + (x2 * 3) + Integer(SCL)).G +
        pRGB24(PtrS * y2 + (x2 * 3) + Integer(SCL)).G) shr 2;
      p.b := (pRGB24(PtrS * y1 + (x1 * 3) + Integer(SCL)).B +
        pRGB24(PtrS * y2 + (x1 * 3) + Integer(SCL)).B +
        pRGB24(PtrS * y1 + (x2 * 3) + Integer(SCL)).B +
        pRGB24(PtrS * y2 + (x2 * 3) + Integer(SCL)).B) shr 2;
    end;
  end;
end;

procedure fxBoxBlur(Src: TBitmap; radius, rep: Integer);
type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;
  TiRGB = packed record B, G, R: Integer; end;
var
  i, j, w, h, x, y, k, n, F, r: Integer;
  SL: array of PRGBArray;
  SL2: PRGBArray;
  LUT: array of TiRGB;
begin
  if radius = 0 then Exit;
  F := Round((1 / ((radius * 2) + 1)) * $10000);
  r := radius - 1;
  w := Src.Width - 1;
  h := Src.Height - 1;
  SetLength(LUT, Max(w, h) + 1 + (radius * 2));
  SetLength(SL, Src.Height);
  SL[0] := (Src.Scanline[0]);
  i := Integer(Src.Scanline[1]) - Integer(SL[0]);
  for j := 1 to Src.Height - 1 do SL[j] := PRGBArray(Integer(SL[j - 1]) + i);
  SL2:= SL[1];
  for y := 0 to h do begin
    for j := 1 to rep do begin
      k := 0;
      for i := -radius to w + radius do begin
        if (i > 0) and (i < w) then k := i;
        LUT[i + radius].R := SL2[k].R + LUT[i + r].R;
        LUT[i + radius].G := SL2[k].G + LUT[i + r].G;
        LUT[i + radius].B := SL2[k].B + LUT[i + r].B;
      end;
      for x := 0 to w do begin
        k := x + radius;
        SL2[x].R := ((LUT[k + radius].R - LUT[k - radius - 1].R) * F) shr 16;
        SL2[x].g := ((LUT[k + radius].G - LUT[k - radius - 1].G) * F) shr 16;
        SL2[x].B := ((LUT[k + radius].B - LUT[k - radius - 1].B) * F) shr 16;
      end;
      SL2 := SL[y];
    end;
  end;
  for x := 0 to w do begin
    for j := 1 to rep do begin
      k := 0;
      for i := -radius to h + radius do begin
        if (i > 0) and (i < h) then k := i;
        LUT[i + radius].R := SL[k][x].R + LUT[i + r].R;
        LUT[i + radius].G := SL[k][x].g + LUT[i + r].G;
        LUT[i + radius].B := SL[k][x].B + LUT[i + r].B;
      end;
      for y := 0 to h do begin
        k := y + radius;
        SL[y][x].R := ((LUT[k + radius].R - LUT[k - radius - 1].R) * F) shr 16;
        SL[y][x].g := ((LUT[k + radius].G - LUT[k - radius - 1].G) * F) shr 16;
        SL[y][x].B := ((LUT[k + radius].B - LUT[k - radius - 1].B) * F) shr 16;
      end;
    end;
  end;
  SetLength(LUT, 0);
end;

procedure TfrmMain.DoTiltShift(Bmp: TBitmap; Pt1, Pt2: TPoint; Render, Save: Boolean);
const
  pi2 = 1.57079633;
var
  s, w, h, i, j, k, sz: Integer;
  bl1, el1, bl2, el2: TPoint;
  bx, by, ex, ey, ang: Integer;
  a: Single;
  apt1, apt2, pt3, pt4, pt5, pt6: TPoint;
  mcr: TRect;
  AImg, Mask: TBitmap;
  x, y, h1, h2, SrcFill, Src2Fill, DstFill: Integer;
  SRow, S2Row, DRow: PRGBArray;
  Alpha: Byte;
begin
  if (not ImgLoaded) then Exit;

  Mask:= TBitmap.Create;
  Mask.PixelFormat:= pf24bit;
  Mask.Width:= Bmp.Width;
  Mask.Height:= Bmp.Height;

  mcr:= Mask.Canvas.ClipRect;
  apt1.X:= MulDiv(PT1.X, Mask.Width, 100);
  apt1.Y:= MulDiv(PT1.Y, Mask.Height, 100);
  apt2.X:= MulDiv(PT2.X, Mask.Width, 100);
  apt2.Y:= MulDiv(PT2.Y, Mask.Height, 100);
  a:= GetAnglePt2FromPt1(aPT1, aPT2);
  sz:= MulDiv(Bmp.Height, tb2.Position, 100) shr 1;
  pt3:= GetPtAtDistAndAngleFromPt(aPT2, -sz, a + pi2);
  pt4:= GetPtAtDistAndAngleFromPt(aPT1, -sz, a + pi2);
  pt5:= GetPtAtDistAndAngleFromPt(aPT2, sz, a + pi2);
  pt6:= GetPtAtDistAndAngleFromPt(aPT1, sz, a + pi2);
  w:= PT4.X - PT3.X;
  h:= PT3.Y - PT4.Y;
  if (pt3.X = pt4.X) or (pt3.Y = pt4.Y) then begin
    if pt3.X = pt4.X then begin
      pt3.Y:= 0; pt4.Y:= Img.Height; pt5.Y:= 0; pt6.Y:= Img.Height;
    end else begin
      pt3.X:= 0; pt4.X:= Img.Width; pt5.X:= 0; pt6.X:= Img.Width;
    end;
    bl1:= pt3; el1:= pt4;
    bl2:= pt5; el2:= pt6;
  end else begin
    i:= (Mask.Height - PT3.Y);
    bl1:= Point(PT3.X - Round((i / h ) * w), Mask.Height);
    el1:= Point(PT4.X + Round((PT4.Y / h ) * w) , 0);
    ClipLine(bl1.X, bl1.Y, el1.X, el1.Y, mcr);
    w:= PT6.X - PT5.X;
    h:= PT5.Y - PT6.Y;
    i:= (Mask.Height - PT5.Y);
    bl2:= Point(PT5.X - Round((i / h ) * w), Mask.Height);
    el2:= Point(PT6.X + Round((PT6.Y / h ) * w) , 0);
    ClipLine(bl2.X, bl2.Y, el2.X, el2.Y, mcr);
  end;
  s:= 10;
  Mask.Canvas.Brush.Color:= clBlack;
  Mask.Canvas.Pen.Color := clBlack;
  Mask.Canvas.Rectangle(mcr);
  Mask.Canvas.Pen.Color := rgb(255, 255, 255);
  if (bl1.X < Mask.Width) and (bl1.Y = 0) then Dec(bl1.Y);
  if (el1.X < Mask.Width) and (el1.Y = 0) then Dec(el1.Y);
  if (bl2.X < Mask.Width) and (bl2.Y = 0) then Dec(bl2.Y);
  if (el2.X < Mask.Width) and (el2.Y = 0) then Dec(el2.Y);
  if (bl1.X = 0) then Dec(bl1.X);
  if (el1.X = 0) then Dec(el1.X);
  if (bl2.X = 0) then Dec(bl2.X);
  if (el2.X = 0) then Dec(el2.X);
  Mask.Canvas.MoveTo(bl1.X, bl1.Y);
  Mask.Canvas.LineTo(el1.X, el1.Y);
  Mask.Canvas.MoveTo(bl2.X, bl2.Y);
  Mask.Canvas.LineTo(el2.X, el2.Y);
  Mask.Canvas.Brush.Color:= rgb(255, 0, 0);
  Mask.Canvas.FloodFill(aPT1.X, aPT1.Y, clBlack, fsSurface);
  fxBoxBlur(Mask, tb1.Position, 1);
  if not Render then begin
    Mask.Canvas.Pen.Color := rgb(0, 0, 255);
    Mask.Canvas.MoveTo(bl1.X, bl1.Y);
    Mask.Canvas.LineTo(el1.X, el1.Y);
    Mask.Canvas.MoveTo(bl2.X, bl2.Y);
    Mask.Canvas.LineTo(el2.X, el2.Y);
  end;
  AImg:= TBitmap.Create;
  AImg.Assign(Bmp);
  if Render then begin
    fxBoxBlur(AImg, tb3.Position, 2);
    fxSplitBlur(AImg, tb4.Position);
  end;
  SRow:= PRGBArray(Bmp.ScanLine[0]);
  SrcFill:= Integer(Bmp.ScanLine[1]) - Integer(SRow);
  S2Row:= PRGBArray(AImg.ScanLine[0]);
  Src2Fill:= Integer(AImg.ScanLine[1]) - Integer(S2Row);
  DRow := PRGBArray(Mask.ScanLine[0]);
  DstFill := Integer(Mask.ScanLine[1]) - Integer(DRow);
  if Render then begin
    for y:= 0 to Mask.Height - 1 do begin
      for x:= 0 to Mask.Width - 1 do begin
        Alpha:= DRow[x].R;
        if Alpha = 255 then DRow[x] := SRow[x] else begin
          DRow[X].R:= Byte(Alpha * (SRow[X].R - S2Row[X].R) shr 8 + S2Row[X].R);
          DRow[X].G:= Byte(Alpha * (SRow[X].G - S2Row[X].G) shr 8 + S2Row[X].G);
          DRow[X].B:= Byte(Alpha * (SRow[X].B - S2Row[X].B) shr 8 + S2Row[X].B);
        end;
      end;
      Inc(Integer(SRow), SrcFill);
      Inc(Integer(S2Row), Src2Fill);
      Inc(Integer(DRow), DstFill);
    end;
    fxBCS(Mask, 255, 20, 357); // Add some contrast and saturation... 20% contrast, 40% saturation...
  end else begin
    for y:= 0 to Mask.Height - 1 do begin
      for x:= 0 to Mask.Width - 1 do begin
        Alpha:= DRow[x].R;
        if Alpha = 255 then DRow[x] := SRow[x] else
        if DRow[X].B = 255 then begin
          DRow[X].R:= 0;
          DRow[X].G:= 255;
          DRow[X].B:= 0;
        end else begin
          DRow[X].R:= Byte(Alpha * (SRow[X].R - 255) shr 8 + 255);
          if chk1.Checked then begin
            DRow[X].G:= SRow[X].G;
            DRow[X].B:= SRow[X].B;
          end else begin
            DRow[X].G:= Byte(Alpha * (SRow[X].G - 0) shr 8 + 0);
            DRow[X].B:= Byte(Alpha * (SRow[X].B - 0) shr 8 + 0);
          end;
        end;
      end;
      Inc(Integer(SRow), SrcFill);
      Inc(Integer(S2Row), Src2Fill);
      Inc(Integer(DRow), DstFill);
    end;
  end;
  if Save then begin
    Bmp.Assign(Mask);
  end else begin
    imgView.Picture.Bitmap.Assign(Mask);
    if not render then begin
      imgView.Picture.Bitmap.Canvas.Ellipse(aPT1.X-s + 1, aPT1.Y-s + 1, aPT1.X+s + 1, aPT1.Y+s + 1);
      imgView.Picture.Bitmap.Canvas.Ellipse(aPT2.X-s + 1, aPT2.Y-s + 1, aPT2.X+s + 1, aPT2.Y+s + 1);
    end;
  end;
  imgView.Refresh;
  AImg.Free;
  Mask.Free;
end;

procedure TfrmMain.MakePreviewImage;
var
  pt: TPoint;
begin
  pt := CalcImgSize(TheImage.Width, TheImage.Height, ImgView.Width, ImgView.Height);
  if (pt.X < 2) or (pt.Y < 2) then  Exit;
  Img.Width := pt.X;
  Img.Height := pt.Y;
  SmoothReSize(TheImage, Img);
  ImgView.Picture.Bitmap.Assign(Img);
  ImgView.Refresh;
end;

procedure TfrmMain.tb2Change(Sender: TObject);
begin
  DoTiltShift(Img, PT1, PT2, False, False);
end;

procedure TfrmMain.btnGoClick(Sender: TObject);
var
  Bmp: TBitmap;
  Jpg: TJPEGImage;
begin
  dlgSaveImg.FileName:= 'TiltShift_' + ExtractFileName(dlgOpenImg.FileName);
  if dlgSaveImg.Execute then begin
    Bmp:= TBitmap.Create;
    Bmp.Assign(TheImage);
    DoTiltShift(Bmp, PT1, PT2, True, True);
    Jpg:= TJPEGImage.Create;
    Jpg.CompressionQuality:= 90;
    Jpg.Assign(Bmp);
    Jpg.SaveToFile(dlgSaveImg.FileName);
    Jpg.Free;
    Bmp.Free;
  end;
  DoTiltShift(Img, PT1, PT2, True, False);
end;

procedure TfrmMain.btnOpenClick(Sender: TObject);
begin
  if dlgOpenImg.Execute then begin
    imgView.Picture.LoadFromFile(dlgOpenImg.Filename);
    TheImage.Assign(imgView.Picture.Graphic);
    TheImage.PixelFormat:= pf24bit;
    imgView.Picture.Bitmap.Assign(TheImage);
    MakePreviewImage;
    PT1.X:= 10;
    PT1.Y:= 50;
    PT2.X:= 90;
    PT2.Y:= 50;
    DoTiltShift(Img, PT1, PT2, False, False);
    ImgLoaded:= True;
    imgView.Enabled:= ImgLoaded;
  end;
end;

procedure TfrmMain.chk1Click(Sender: TObject);
begin
  DoTiltShift(Img, PT1, PT2, False, False);
end;

procedure TfrmMain.FormCreate(Sender: TObject);
var
  i: Integer;
  C: ColorRef;
  B, G, R: Byte;
begin
  ImgLoaded:= False;
  TheImage:= TBitmap.Create;
  Img:= TBitmap.Create;
  Img.PixelFormat:= pf24bit;
  // Setup selection color
  C := ColorToRGB(clHighlight);
  B := Byte(C shr 16);
  G := Byte(C shr 8);
  R := Byte(C);
  for i := 0 to 255 do begin
    aR[i] := (R - i) shr 1 + i;
    aG[i] := (G - i) shr 1 + i;
    aB[i] := (B - i) shr 1 + i;
  end;
  imgView.Picture.Bitmap.Width:= 1;
  imgView.Picture.Bitmap.Height:= 1;
  imgView.Picture.Bitmap.Canvas.Pixels[0, 0]:= $001A1A1A;
  LY:= -1;
  PT1.X:= 20;
  PT1.Y:= imgView.Picture.Bitmap.Height shr 1;
  PT2.X:= imgView.Picture.Bitmap.Width - 21;
  PT2.Y:= imgView.Picture.Bitmap.Height shr 1;
  DoNow:= False;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  Img.Free;
  TheImage.Free;
end;

procedure TfrmMain.FormResize(Sender: TObject);
begin
  if ImgLoaded then MakePreviewImage;
end;

procedure TfrmMain.imgViewMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if not ImgLoaded then Exit;
  CP:= -1;
  X:= X - ((imgView.Width - Img.Width) shr 1);
  Y:= Y - ((imgView.Height - Img.Height) shr 1);
  if x < 0 then x:= 0 else if x > Img.Width - 1 then x:= Img.Width - 1;
  if y < 0 then y:= 0 else if y > Img.Height - 1 then y:= Img.Height - 1;
  X:= MulDiv(X, 100, Img.Width);
  Y:= MulDiv(Y, 100, Img.Height);
  if Distance(X, Y, PT1.X, PT1.Y) < Distance(X, Y, PT2.X, PT2.Y) then begin
    PT1.X:= X; PT1.Y:= Y;
    CP:= 1;
  end else begin
    PT2.X:= X; PT2.Y:= Y;
    CP:= 2;
  end;
end;

procedure TfrmMain.imgViewMouseEnter(Sender: TObject);
begin
  DoTiltShift(Img, PT1, PT2, False, False);
end;

procedure TfrmMain.imgViewMouseLeave(Sender: TObject);
begin
  DoTiltShift(Img, PT1, PT2, True, False);
end;

procedure TfrmMain.imgViewMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if (not ImgLoaded) then Exit;
  if (ssLeft in Shift) and (CP <> -1) then begin
    X:= X - ((imgView.Width - Img.Width) shr 1);
    Y:= Y - ((imgView.Height - Img.Height) shr 1);
    if x < 2 then x:= 2 else if x > Img.Width - 3 then x:= Img.Width - 3;
    if y < 2 then y:= 2 else if y > Img.Height - 3 then y:= Img.Height - 3;
    X:= MulDiv(X, 100, Img.Width);
    Y:= MulDiv(Y, 100, Img.Height);
    if CP = 1 then PT1:= Point(X, Y) else PT2:= Point(X, Y);
    DoTiltShift(Img, PT1, PT2, False, False);
  end;
end;

procedure TfrmMain.imgViewMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  CP:= -1;
end;

end.
