unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, Jpeg, Math;

type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte;
  end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;

  TfrmMain = class(TForm)
    pnlTop: TPanel;
    statMain: TStatusBar;
    imgView: TImage;
    dlgOpenImg: TOpenDialog;
    btnOpen: TButton;
    btnGo: TButton;
    procedure btnOpenClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure imgViewMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure imgViewMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure imgViewMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure imgViewMouseLeave(Sender: TObject);
  private
    procedure MakePreviewImage;
    procedure AlphaSelection;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

var
  TheImage, Img: TBitmap;
  P1, P2, LY: Integer;
  aR, aG, aB: array[0..255] of Byte;
  ImgLoaded: Boolean;
  DoNow: Boolean;

function CalcImgSize(w, h, tw, th: integer): TPoint;
begin
  Result.X := 0;
  Result.Y := 0;
  if (w < tw) and (h < th) then begin
    Result.X := w; Result.Y := h;
  end
  else if (w = 0) or (h = 0) then Exit
  else begin
    if w > h then begin
      if w < tw then tw := w;
      Result.X := tw;   Result.Y := Trunc(tw * h / w);
      if Result.Y > th then begin
        Result.Y := th; Result.X := Trunc(th * w / h);
      end;
    end else begin
      if h < th then th := h;
      Result.Y := th;   Result.X := Trunc(th * w / h);
      if Result.X > tw then begin
        Result.X := tw; Result.Y := Trunc(tw * h / w);
      end;
    end;
  end;
end;

procedure SmoothReSize(Bmp, Dest: tBitmap);
var
  x, y, R, G, B, PSrc1, PSrc2, SrcFill, PDst, DstFill: integer;
  Src : TBitmap;
begin
  Src:= TBitmap.Create;
  Src.PixelFormat:= pf24bit;
  Src.Width:= Dest.Width shl 1;
  Src.Height:= Dest.Height shl 1;
  Src.Canvas.StretchDraw(Src.Canvas.ClipRect, bmp);
  PSrc1:= Integer(Src.ScanLine[0]);
  PSrc2:= Integer(Src.ScanLine[1]);
  SrcFill := (PSrc2 - PSrc1);
  SrcFill := SrcFill + (SrcFill - Src.Width * 3);
  PDst := Integer(Dest.ScanLine[0]);
  DstFill := (Integer(Dest.ScanLine[1]) - PDst) - (Dest.Width * 3);
  for y:= 0 to Dest.Height - 1 do begin
    for x:= 0 to Dest.Width - 1 do begin
      B:= Pbyte(PSrc1)^;
      G:= PByte(PSrc1 + 1)^;
      R:= PByte(PSrc1 + 2)^;
      B:= B + PByte(PSrc1 + 3)^;
      G:= G + PByte(PSrc1 + 4)^;
      R:= R + PByte(PSrc1 + 5)^;
      B:= B + PByte(PSrc2)^;
      G:= G + PByte(PSrc2 + 1)^;
      R:= R + PByte(PSrc2 + 2)^;
      B:= B + PByte(PSrc2 + 3)^;
      G:= G + PByte(PSrc2 + 4)^;
      R:= R + PByte(PSrc2 + 5)^;
      Inc(PSrc1, 6);
      Inc(PSrc2, 6);
      PByte(PDst)^:= B shr 2;
      PByte(PDst + 1)^:= G shr 2;
      PByte(PDst + 2)^:= R shr 2;
      Inc(PDst, 3);
    end;
    Inc(PSrc1, SrcFill);
    Inc(PSrc2, SrcFill);
    Inc(PDst, DstFill);
  end;
  Src.Free;
end;

procedure fxBCS(Dest: TBitmap; B, C, S: integer);
// Brightness values [0..512] def 255
// Contrast values [-100..100] def 0
// Saturation values [0..512] def 255
var
  x, y, i, k, v, pv, ci1, ci2, ci3, alpha, s1, s2, SLDest, DestOff: Integer;
  a: double;
  p: PRGB24;
  color: TRGB24;
  BrightnesLut, ContrastLut, BCLut: array[0..255] of byte;
  SaturationLut: record
    Grays: array[0..767] of Integer; Alpha: array[Byte] of Word;
  end;
begin
  if c = 100 then c := 99;
  for i := 0 to 255 do begin
    if c > 0 then a := 1 / cos(c * (PI / 200))
    else          a := 1 * cos(c * (3.1416 / 200));
    v := Round(a * (i - 170) + 170);
    if v > 255 then v := 255 else if v < 0 then v := 0;
    ContrastLut[i] := v;
  end;
  for i := 0 to 255 do begin
    alpha := b;
    k := 256 - alpha;
    v := (k + alpha * i) div 256;
    if v < 0 then v := 0 else if v > 255 then v := 255;
    BCLut[i] := ContrastLut[v];
  end;
  x := 0;
  for i := 1 to 256 do SaturationLut.Alpha[i - 1] := (i * s) shr 8;
  for i := 0 to 255 do begin
    y := i - SaturationLut.Alpha[i];
    SaturationLut.Grays[x] := y; Inc(x);
    SaturationLut.Grays[x] := y; Inc(x);
    SaturationLut.Grays[x] := y; Inc(x);
  end;
  SLDest := Integer(dest.ScanLine[0]);
  DestOff := Integer(dest.ScanLine[1]) - SLDest;
  for y := 0 to Dest.Height - 1 do begin
    for x := 0 to Dest.Width - 1 do begin
      color := pRGBArray(SLDest)[x];
      v := color.R + color.G + color.B;
      ci1 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.B];
      ci2 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.G];
      ci3 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.R];
      if ci1 < 0 then ci1 := 0 else if ci1 > 255 then ci1 := 255;
      if ci2 < 0 then ci2 := 0 else if ci2 > 255 then ci2 := 255;
      if ci3 < 0 then ci3 := 0 else if ci3 > 255 then ci3 := 255;
      PRGBArray(SLDest)[x].B := BCLut[ci1];
      PRGBArray(SLDest)[x].G := BCLut[ci2];
      PRGBArray(SLDest)[x].R := BCLut[ci3];
    end;
    inc(SLDest, DestOff);
  end;
end;

procedure fxBoxBlur(Src: TBitmap; radius, rep: Integer);
type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;
  TiRGB = packed record B, G, R: Integer; end;
var
  i, j, w, h, x, y, k, n, F, r: Integer;
  SL: array of PRGBArray;
  SL2: PRGBArray;
  LUT: array of TiRGB;
begin
  if radius = 0 then Exit;
  F := Round((1 / ((radius * 2) + 1)) * $10000);
  r := radius - 1;
  w := Src.Width - 1;
  h := Src.Height - 1;
  SetLength(LUT, Max(w, h) + 1 + (radius * 2));
  SetLength(SL, Src.Height);
  SL[0] := (Src.Scanline[0]);
  i := Integer(Src.Scanline[1]) - Integer(SL[0]);
  for j := 1 to Src.Height - 1 do SL[j] := PRGBArray(Integer(SL[j - 1]) + i);
  SL2:= SL[1];
  for y := 0 to h do begin
    for j := 1 to rep do begin
      k := 0;
      for i := -radius to w + radius do begin
        if (i > 0) and (i < w) then k := i;
        LUT[i + radius].R := SL2[k].R + LUT[i + r].R;
        LUT[i + radius].G := SL2[k].G + LUT[i + r].G;
        LUT[i + radius].B := SL2[k].B + LUT[i + r].B;
      end;
      for x := 0 to w do begin
        k := x + radius;
        SL2[x].R := ((LUT[k + radius].R - LUT[k - radius - 1].R) * F) shr 16;
        SL2[x].g := ((LUT[k + radius].G - LUT[k - radius - 1].G) * F) shr 16;
        SL2[x].B := ((LUT[k + radius].B - LUT[k - radius - 1].B) * F) shr 16;
      end;
      SL2 := SL[y];
    end;
  end;
  for x := 0 to w do begin
    for j := 1 to rep do begin
      k := 0;
      for i := -radius to h + radius do begin
        if (i > 0) and (i < h) then k := i;
        LUT[i + radius].R := SL[k][x].R + LUT[i + r].R;
        LUT[i + radius].G := SL[k][x].g + LUT[i + r].G;
        LUT[i + radius].B := SL[k][x].B + LUT[i + r].B;
      end;
      for y := 0 to h do begin
        k := y + radius;
        SL[y][x].R := ((LUT[k + radius].R - LUT[k - radius - 1].R) * F) shr 16;
        SL[y][x].g := ((LUT[k + radius].G - LUT[k - radius - 1].G) * F) shr 16;
        SL[y][x].B := ((LUT[k + radius].B - LUT[k - radius - 1].B) * F) shr 16;
      end;
    end;
  end;
  SetLength(LUT, 0);
end;

procedure fxTiltShift(Bmp: TBitmap; P1, P2, Fade: Integer);
var
  Img: TBitmap;
  i, x, y, a, h1, h2, SrcFill, DstFill: Integer;
  SRow, DRow: PRGBArray;
  Alpha: array of Byte;
begin
  Img:= TBitmap.Create;
  Img.Assign(Bmp);
  fxBoxBlur(Img, 3, 3);
  h1:= MulDiv(P1, Img.Height, 100);
  h2:= MulDiv(P2, Img.Height, 100);
  if h1 < 0 then h1:= 0 else if h1 > Img.Height - 1 then h1:= Img.Height - 1;
  if h2 < 0 then h2:= 0 else if h2 > Img.Height - 1 then h2:= Img.Height - 1;
  SetLength(Alpha, Img.Height);
  a:= 255;
  for i := 0 to Img.Height - 1 do begin // Set alpha values
    if i = h1 then a:= 0;
    if i = h2 then a:= 255;
    Alpha[i]:= a;
  end;
  for i := 0 to Fade - 1 do begin // Blend the edges
    a:= MulDiv(i, 255, Fade); //Round(25.5 * i);
    Alpha[h1 + i]:= 255 - a;
    if (h2 + i) < Img.Height then
      Alpha[h2 + i]:= a;
  end;
  SRow:= PRGBArray(Img.ScanLine[0]);
  SrcFill := Integer(Img.ScanLine[1]) - Integer(SRow);
  DRow := PRGBArray(bmp.ScanLine[0]);
  DstFill := Integer(bmp.ScanLine[1]) - Integer(DRow);
  for y:= 0 to Img.Height - 1 do begin // Copy image 1 to image 2 using alpha value values
    a:= Alpha[y];
    for x:= 0 to Img.Width - 1 do begin
       if a > 0 then if a = 255 then DRow[x]:= SRow[x]
       else begin
         DRow[X].R:= Byte(a * (SRow[X].R - DRow[X].R) shr 8 + DRow[X].R);
         DRow[X].G:= Byte(a * (SRow[X].G - DRow[X].G) shr 8 + DRow[X].G);
         DRow[X].B:= Byte(a * (SRow[X].B - DRow[X].B) shr 8 + DRow[X].B);
       end;
    end;
    Inc(Integer(SRow), SrcFill);
    Inc(Integer(DRow), DstFill);
  end;
  fxBCS(bmp, 255, 20, 357); // Add some contrast and saturation... 20% contrast, 40% saturation...
  Img.Free;
end;

procedure TfrmMain.AlphaSelection;
var
  bmp: TBitmap;
  a, i, x, y, h1, h2, Row, DstFill: Integer;
  p: PRGB24;
  Alpha: array of Byte;
begin
  bmp:= TBitmap.Create;
  bmp.Assign(Img);
  h1:= MulDiv(P1, bmp.Height, 100);
  h2:= MulDiv(P2, bmp.Height, 100);
  if h1 < 0 then h1:= 0 else if h1 > bmp.Height - 1 then h1:= bmp.Height - 1;
  if h2 < 0 then h2:= 0 else if h2 > bmp.Height - 1 then h2:= bmp.Height - 1;
  // Set alpha values
  SetLength(Alpha, bmp.Height);
  a:= 255;
  for i := 0 to bmp.Height - 1 do begin
    if i = h1 then a:= 0;
    if i = h2 then a:= 255;
    Alpha[i]:= a;
  end;
  Row := Integer(bmp.ScanLine[0]);
  DstFill := Integer(bmp.ScanLine[1]) - Row;
  for y:= 0 to bmp.Height - 1 do begin
    p:= PRGB24(Row);
    a:= Alpha[y];
    for x:= 0 to bmp.Width - 1 do begin
      if a = 255 then begin p.R:= aR[p.R]; p.G:= aG[p.G]; p.B:= aB[p.B]; end;
      Inc(p);
    end;
    Inc(Row, DstFill);
  end;
  imgView.Picture.Bitmap.Assign(bmp);
  imgView.Refresh;
  SetLength(Alpha, 0);
  bmp.Free;
end;

procedure TfrmMain.MakePreviewImage;
var
  pt: TPoint;
begin
  pt := CalcImgSize(TheImage.Width, TheImage.Height, ImgView.Width, ImgView.Height);
  if (pt.X < 2) or (pt.Y < 2) then  Exit;
  Img.Width := pt.X;
  Img.Height := pt.Y;
  SmoothReSize(TheImage, Img);
  ImgView.Picture.Bitmap.Assign(Img);
  ImgView.Refresh;
  LY:= -1;
end;

procedure TfrmMain.btnOpenClick(Sender: TObject);
begin
  if dlgOpenImg.Execute then begin
    imgView.Picture.LoadFromFile(dlgOpenImg.Filename);
    TheImage.Assign(imgView.Picture.Graphic);
    TheImage.PixelFormat:= pf24bit;
    imgView.Picture.Bitmap.Assign(TheImage);
    MakePreviewImage;
    ImgLoaded:= True;
    imgView.Enabled:= ImgLoaded;
  end;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
var
  i: Integer;
  C: ColorRef;
  B, G, R: Byte;
begin
  ImgLoaded:= False;
  TheImage:= TBitmap.Create;
  Img:= TBitmap.Create;
  Img.PixelFormat:= pf24bit;
  // Setup selection color
  C := ColorToRGB(clHighlight);
  B := Byte(C shr 16);
  G := Byte(C shr 8);
  R := Byte(C);
  for i := 0 to 255 do begin
    aR[i] := (R - i) shr 1 + i;
    aG[i] := (G - i) shr 1 + i;
    aB[i] := (B - i) shr 1 + i;
  end;
  imgView.Picture.Bitmap.Width:= 1;
  imgView.Picture.Bitmap.Height:= 1;
  imgView.Picture.Bitmap.Canvas.Pixels[0, 0]:= $00C4C4C4;
  LY:= -1;
  DoNow:= False;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  Img.Free;
  TheImage.Free;
end;

procedure TfrmMain.FormResize(Sender: TObject);
begin
  if ImgLoaded then MakePreviewImage;
end;

procedure TfrmMain.imgViewMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  P1:= MulDiv(100, Y - ((imgView.Height - Img.Height) shr 1), Img.Height);
  DoNow:= (ssLeft in Shift);
end;

procedure TfrmMain.imgViewMouseLeave(Sender: TObject);
begin
  if (LY <> -1) and (ImgLoaded) then begin
    imgView.Canvas.MoveTo(0, LY); imgView.Canvas.LineTo(imgView.Width - 1, LY);
  end;
  LY:= -1;
end;

procedure TfrmMain.imgViewMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  Y:= Y - ((imgView.Height - Img.Height) shr 1);
  if (not (ssLeft in Shift)) and (ImgLoaded) then begin
    imgView.Canvas.Pen.Color := rgb(255,255,255);
    imgView.Canvas.Pen.Mode:= pmXor;
    imgView.Canvas.Pen.Style:= psSolid;
    if (LY <> -1) then begin
      imgView.Canvas.MoveTo(0, LY); imgView.Canvas.LineTo(imgView.Width - 1, LY);
    end;
    imgView.Canvas.MoveTo(0, Y); imgView.Canvas.LineTo(imgView.Width - 1, Y);
    LY:= Y;
    imgView.Refresh;
  end else if (ImgLoaded) and (DoNow) then begin
    LY:= -1;
    P2:= MulDiv(100, Y, Img.Height);
    AlphaSelection;
    Caption:= 'First percent: ' + IntToStr(p1) + ', Second percent: ' + IntToStr(p2);
  end;
end;

procedure TfrmMain.imgViewMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if (ImgLoaded) and (DoNow) then begin
    MakePreviewImage;
    fxTiltShift(Img, P1, P2, 50);
    imgView.Picture.Bitmap.Assign(Img);
    imgView.Refresh;
  end;
  DoNow:= False;
end;

end.
